package utils

import (
	"encoding/json"
	"errors"
	"fmt"
	"time"

	"cloud.o-forge.io/core/oc-lib/dbs"
	"cloud.o-forge.io/core/oc-lib/dbs/mongo"
	"cloud.o-forge.io/core/oc-lib/tools"
	"github.com/go-playground/validator/v10"
	"github.com/google/uuid"
	"github.com/rs/zerolog"
	mgb "go.mongodb.org/mongo-driver/mongo"
)

// single instance of the validator used in every model Struct to validate the fields
var validate = validator.New(validator.WithRequiredStructEnabled())

/*
* AbstractObject is a struct that represents the basic fields of an object
* it defines the object id and name
* every data in base root model should inherit from this struct (only exception is the ResourceModel)
 */
type AbstractObject struct {
	UUID           string    `json:"id,omitempty" bson:"id,omitempty" validate:"required"`
	Name           string    `json:"name,omitempty" bson:"name,omitempty" validate:"required"`
	UpdateDate     time.Time `json:"update_date" bson:"update_date"`
	LastPeerWriter string    `json:"last_peer_writer" bson:"last_peer_writer"`
}

func (r *AbstractObject) GenerateID() {
	if r.UUID == "" {
		r.UUID = uuid.New().String()
	}
}

// GetID implements ShallowDBObject.
func (ao AbstractObject) GetID() string {
	return ao.UUID
}

// GetName implements ShallowDBObject.
func (ao AbstractObject) GetName() string {
	return ao.Name
}

func (ao *AbstractObject) UpToDate() {
	ao.UpdateDate = time.Now()
	// ao.LastPeerWriter, _ = static.GetMyLocalJsonPeer()
}

func (ao *AbstractObject) VerifyAuth(peerID string, groups []string) bool {
	return true
}

func (ao *AbstractObject) GetObjectFilters(search string) *dbs.Filters {
	return &dbs.Filters{
		Or: map[string][]dbs.Filter{ // filter by name if no filters are provided
			"abstractobject.name": {{Operator: dbs.LIKE.String(), Value: search}},
		}}
}

func (dma *AbstractObject) Deserialize(j map[string]interface{}, obj DBObject) DBObject {
	b, err := json.Marshal(j)
	if err != nil {
		return nil
	}
	json.Unmarshal(b, obj)
	return obj
}

func (dma *AbstractObject) Serialize(obj DBObject) map[string]interface{} {
	var m map[string]interface{}
	b, err := json.Marshal(obj)
	if err != nil {
		return nil
	}
	json.Unmarshal(b, &m)
	return m
}

type AbstractAccessor struct {
	Logger zerolog.Logger    // Logger is the logger of the accessor, it's a specilized logger for the accessor
	Type   tools.DataType    // Type is the data type of the accessor
	Caller *tools.HTTPCaller // Caller is the http caller of the accessor (optionnal) only need in a peer connection
	PeerID string            // PeerID is the id of the peer
	Groups []string          // Groups is the list of groups that can access the accessor

	ResourceModelAccessor Accessor
}

func (dma *AbstractAccessor) GetPeerID() string {
	return dma.PeerID
}
func (dma *AbstractAccessor) GetGroups() []string {
	return dma.Groups
}
func (dma *AbstractAccessor) GetLogger() *zerolog.Logger {
	return &dma.Logger
}

func (dma *AbstractAccessor) VerifyAuth() string {
	return ""
}

func (dma *AbstractAccessor) GetType() tools.DataType {
	return dma.Type
}

func (dma *AbstractAccessor) GetCaller() *tools.HTTPCaller {
	return dma.Caller
}

// GenericLoadOne loads one object from the database (generic)
func GenericStoreOne(data DBObject, a Accessor) (DBObject, int, error) {
	data.GenerateID()
	f := dbs.Filters{
		Or: map[string][]dbs.Filter{
			"abstractresource.abstractobject.name": {{
				Operator: dbs.LIKE.String(),
				Value:    data.GetName(),
			}},
			"abstractobject.name": {{
				Operator: dbs.LIKE.String(),
				Value:    data.GetName(),
			}},
		},
	}
	if !data.VerifyAuth(a.GetPeerID(), a.GetGroups()) {
		return nil, 403, errors.New("You are not allowed to access this collaborative area")
	}
	if cursor, _, _ := a.Search(&f, ""); len(cursor) > 0 {
		return nil, 409, errors.New(a.GetType().String() + " with name " + data.GetName() + " already exists")
	}
	err := validate.Struct(data)
	if err != nil {
		return nil, 422, err
	}
	id, code, err := mongo.MONGOService.StoreOne(data, data.GetID(), a.GetType().String())
	if err != nil {
		a.GetLogger().Error().Msg("Could not store " + data.GetName() + " to db. Error: " + err.Error())
		return nil, code, err
	}
	return a.LoadOne(id)
}

// GenericLoadOne loads one object from the database (generic)
func GenericDeleteOne(id string, a Accessor) (DBObject, int, error) {
	res, code, err := a.LoadOne(id)
	if err != nil {
		a.GetLogger().Error().Msg("Could not retrieve " + id + " to db. Error: " + err.Error())
		return nil, code, err
	}
	if !res.VerifyAuth(a.GetPeerID(), a.GetGroups()) {
		return nil, 403, errors.New("You are not allowed to access this collaborative area")
	}
	_, code, err = mongo.MONGOService.DeleteOne(id, a.GetType().String())
	if err != nil {
		a.GetLogger().Error().Msg("Could not delete " + id + " to db. Error: " + err.Error())
		return nil, code, err
	}
	return res, 200, nil
}

// GenericLoadOne loads one object from the database (generic)
// json expected in entry is a flatted object no need to respect the inheritance hierarchy
func GenericUpdateOne(set DBObject, id string, a Accessor, new DBObject) (DBObject, int, error) {
	r, c, err := a.LoadOne(id)
	if err != nil {
		return nil, c, err
	}
	if !r.VerifyAuth(a.GetPeerID(), a.GetGroups()) {
		return nil, 403, errors.New("You are not allowed to access this collaborative area")
	}
	change := set.Serialize(set) // get the changes
	loaded := r.Serialize(r)     // get the loaded object

	for k, v := range change { // apply the changes, with a flatten method
		loaded[k] = v
	}
	id, code, err := mongo.MONGOService.UpdateOne(new.Deserialize(loaded, new), id, a.GetType().String())
	if err != nil {
		a.GetLogger().Error().Msg("Could not update " + id + " to db. Error: " + err.Error())
		return nil, code, err
	}
	return a.LoadOne(id)
}

func GenericLoadOne[T DBObject](id string, f func(DBObject) (DBObject, int, error), a Accessor) (DBObject, int, error) {
	var data T
	res_mongo, code, err := mongo.MONGOService.LoadOne(id, a.GetType().String())
	if err != nil {
		a.GetLogger().Error().Msg("Could not retrieve " + id + " from db. Error: " + err.Error())
		return nil, code, err
	}
	res_mongo.Decode(&data)
	if !data.VerifyAuth(a.GetPeerID(), a.GetGroups()) {
		return nil, 403, errors.New("You are not allowed to access this collaborative area")
	}
	return f(data)
}

func genericLoadAll[T DBObject](res *mgb.Cursor, code int, err error, f func(DBObject) ShallowDBObject, a Accessor) ([]ShallowDBObject, int, error) {
	objs := []ShallowDBObject{}
	var results []T
	if err != nil {
		a.GetLogger().Error().Msg("Could not retrieve any from db. Error: " + err.Error())
		return nil, code, err
	}
	if err = res.All(mongo.MngoCtx, &results); err != nil {
		return nil, 404, err
	}
	fmt.Println("results 1", res, results)
	for _, r := range results {
		fmt.Println("results", r.VerifyAuth(a.GetPeerID(), a.GetGroups()))
		if !r.VerifyAuth(a.GetPeerID(), a.GetGroups()) {
			continue
		}
		objs = append(objs, f(r))
	}
	return objs, 200, nil
}

func GenericLoadAll[T DBObject](f func(DBObject) ShallowDBObject, wfa Accessor) ([]ShallowDBObject, int, error) {
	res_mongo, code, err := mongo.MONGOService.LoadAll(wfa.GetType().String())
	fmt.Println("res_mongo", res_mongo)
	return genericLoadAll[T](res_mongo, code, err, f, wfa)
}

func GenericSearch[T DBObject](filters *dbs.Filters, search string, defaultFilters *dbs.Filters,
	f func(DBObject) ShallowDBObject, wfa Accessor) ([]ShallowDBObject, int, error) {
	if (filters == nil || len(filters.And) == 0 || len(filters.Or) == 0) && search != "" {
		filters = defaultFilters
	}
	res_mongo, code, err := mongo.MONGOService.Search(filters, wfa.GetType().String())
	return genericLoadAll[T](res_mongo, code, err, f, wfa)
}

// GenericLoadOne loads one object from the database (generic)
// json expected in entry is a flatted object no need to respect the inheritance hierarchy
func GenericRawUpdateOne(set DBObject, id string, a Accessor) (DBObject, int, error) {
	id, code, err := mongo.MONGOService.UpdateOne(set, id, a.GetType().String())
	if err != nil {
		a.GetLogger().Error().Msg("Could not update " + id + " to db. Error: " + err.Error())
		return nil, code, err
	}
	return a.LoadOne(id)
}
